# Generated by Django 2.2.16 on 2020-11-20 14:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0026_deprecation_per_repository'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tag',
            name='name',
            field=models.CharField(editable=False, max_length=64, unique=True),
        ),
    ]
